/*
 * Copyright (c) 2018-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri Dec  7 2018
 */
#include "acgps.h"

#include <err.h>
#include <float.h>
#include <math.h>

#include <Eigen/Core>
#include <iostream>

#include "codels.h"


/* --- UTC ----------------------------------------------------------------- */

/* From:
 * http://pubs.opengroup.org/
 *                      onlinepubs/9699919799/basedefs/V1_chap04.html#tag_04_16
 * https://en.wikipedia.org/wiki/Unix_time
 */

namespace {
  const unsigned int mday[12] = {
    0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334
  };

  const unsigned int gpsepoch = 315964800; /* Jan 6 1980 00:00 UTC */

  const int min_sec = 60;
  const int hour_sec = min_sec * 60;
  const int day_sec = hour_sec * 24;
  const int week_sec = day_sec * 7;
  const int year_sec = day_sec * 365;

  const int sec_msec = 1000;
  const int sec_usec = sec_msec * 1000;
  const int sec_nsec = sec_usec * 1000;
}

or_time_ts
gpsconv_utcmktime(struct tm *tm, int32_t nano)
{
  or_time_ts ts;

  /* clamp month */
  if (tm->tm_mon > 11) {
    tm->tm_year += tm->tm_mon / 12;
    tm->tm_mon = tm->tm_mon % 12;
  } else if (tm->tm_mon < 0) {
    tm->tm_year += tm->tm_mon / 12 - 1;
    tm->tm_mon = (tm->tm_mon % 12 + 12) % 12;
  }

  /* yday */
  tm->tm_yday = tm->tm_mday - 1;
  tm->tm_yday += mday[tm->tm_mon];
  if (tm->tm_mon > 1)
    tm->tm_yday += /* leap year */
      !(tm->tm_year % 4)
      - !(tm->tm_year % 100)
      + !((tm->tm_year + 300) % 400);

  /* sec */
  ts.sec = tm->tm_sec;
  ts.sec += tm->tm_min * min_sec;
  ts.sec += tm->tm_hour * hour_sec;
  ts.sec += tm->tm_yday * day_sec;
  ts.sec += (tm->tm_year - 70) * year_sec;
  ts.sec += /* leap years ... */
    ((tm->tm_year - 69)/4       /* ... from 1973 */
     - (tm->tm_year - 1)/100    /* ... from 2001 */
     + (tm->tm_year + 299)/400) /* ... from 2001 */
    * day_sec;

  /* nsec */
  if (nano >= sec_nsec) {
    ts.sec += nano / sec_nsec;
    ts.nsec = nano % sec_nsec;
  } else if (nano < 0) {
    ts.sec += nano / sec_nsec - 1;
    ts.nsec = (nano % sec_nsec + sec_nsec) % sec_nsec;
  } else
    ts.nsec = nano;

  return ts;
}

or_time_ts
gpsconv_gpsutc(uint16_t week, uint32_t ms, int16_t leaps)
{
  or_time_ts ts;

  ts.sec = gpsepoch + week * week_sec + ms / sec_msec + leaps;
  ts.nsec = (ms % sec_msec) * sec_usec;
  return ts;
}


/* --- gpsconv_utcts ------------------------------------------------------- */

/** Implements Olson, Edwin. "A passive solution to the sensor synchronization
 * problem." International conference on Intelligent Robots and Systems (IROS),
 * 2010 IEEE/RSJ */
or_time_ts
gpsconv_utcts(or_time_ts utc, struct timeval rts, double period, double tdelay)
{
  static double offset = -DBL_MAX;
  double ds;
  int f;

  /* substract transmission delay from reception timestamp */
  f = floor(tdelay);
  rts.tv_sec -= f;
  rts.tv_usec -= sec_usec * (tdelay - f);
  if (rts.tv_usec < 0) {
    rts.tv_sec--;
    rts.tv_usec += sec_usec;
  }

  /* update offset assuming a 1ppm local clock drift */
  offset -= 1./1e6 * period;
  ds = utc.sec - rts.tv_sec;
  ds += utc.nsec / (double)sec_nsec - rts.tv_usec / (double)sec_usec;
  if (ds > offset)
    offset = ds;

  /* return local timestamp */
  f = floor(-offset);
  utc.sec += f;
  utc.nsec += sec_nsec * (-offset - f);
  if (utc.nsec >= sec_nsec) {
    utc.sec++;
    utc.nsec -= sec_nsec;
  }
  return utc;
}


/* --- WGS84 --------------------------------------------------------------- */

/* From:
 *  https://en.wikipedia.org/wiki/World_Geodetic_System#WGS84
 *  https://en.wikipedia.org/wiki/Geographic_coordinate_conversion
 */

namespace {
  /* ellipsoid definition */
  const double wgs84_a = 6378137.0; /* meter */
  const double wgs84_f = 1. / 298.257223563;

  /* additional computed values */
  const double wgs84_b = wgs84_a * (1. - wgs84_f);
  const double wgs84_e2 = 1. - wgs84_b / wgs84_a * wgs84_b / wgs84_a;
  const double wgs84_ep2 = wgs84_a / wgs84_b * wgs84_a / wgs84_b - 1.;
}


/* --- gpsconv_setref ------------------------------------------------------ */

namespace {
  namespace enuref {
    Eigen::Matrix3d m;
    Eigen::Vector3d r;
  }
}

void
gpsconv_setref(const gps_llh_s *llh)
{
  double lat = llh->latitude * M_PI/180.;
  double lon = llh->longitude * M_PI/180.;
  double sin_lat = sin(lat), cos_lat = cos(lat);
  double sin_lon = sin(lon), cos_lon = cos(lon);
  gps_ecef_s ecef;

  gpsconv_llhecef(llh, &ecef);

  enuref::r <<
    ecef.x,
    ecef.y,
    ecef.z;

  enuref::m <<
    - sin_lon,               cos_lon,             0.,
    - sin_lat * cos_lon,   - sin_lat * sin_lon,   cos_lat,
    + cos_lat * cos_lon,     cos_lat * sin_lon,   sin_lat;
}


/* --- gpsconv_llhecef ----------------------------------------------------- */

void
gpsconv_llhecef(const gps_llh_s *llh, gps_ecef_s *ecef)
{
  double lat = llh->latitude * M_PI/180.;
  double lon = llh->longitude * M_PI/180.;
  double sin_lat = sin(lat), cos_lat = cos(lat);
  double sin_lon = sin(lon), cos_lon = cos(lon);

  double N = wgs84_a / sqrt(1. - wgs84_e2 * sin_lat * sin_lat);

  ecef->x = (N + llh->height) * cos_lat * cos_lon;
  ecef->y = (N + llh->height) * cos_lat * sin_lon;
  ecef->z = ((1. - wgs84_e2) * N + llh->height) * sin_lat;
}


/* --- gpsconv_ecefllh ----------------------------------------------------- */

/* from
 * http://www.nalresearch.com/files/Standard%20Modems/A3LA-XG/
 * A3LA-XG%20SW%20Version%201.0.0/GPS%20Technical%20Documents/
 * GPS.G1-X-00006%20(Datum%20Transformations).pdf
 */
void
gpsconv_ecefllh(const gps_ecef_s *ecef, gps_llh_s *llh)
{
  double p = hypot(ecef->x, ecef->y);
  double th = atan2(ecef->z * wgs84_a, p * wgs84_b);

  double sin_th = sin(th), cos_th = cos(th);

  double lon = atan2(ecef->y, ecef->x);
  double lat = atan2(
    ecef->z + wgs84_ep2 * wgs84_b * sin_th * sin_th * sin_th,
    p - wgs84_e2 * wgs84_a * cos_th * cos_th * cos_th);

  double sin_lat = sin(lat), cos_lat = cos(lat);

  double N = wgs84_a / sqrt(1. - wgs84_e2 * sin_lat * sin_lat);

  double h = p / cos_lat - N;

  llh->latitude = lat * 180./M_PI;
  llh->longitude = lon * 180./M_PI;
  llh->height = h;
}


/* --- gpsconv_ecefenu ----------------------------------------------------- */

void
gpsconv_ecefenu(const gps_ecef_s *ecef, gps_enu_s *enu)
{
  Eigen::Vector3d enu_;

  enu_ = enuref::m * (Eigen::Vector3d(ecef->x, ecef->y, ecef->z) - enuref::r);

  enu->east = enu_(0);
  enu->north = enu_(1);
  enu->up = enu_(2);
}

void
gpsconv_decefenu(const gps_ecef_s *ecef, gps_enu_s *enu)
{
  Eigen::Vector3d enu_;

  enu_ = enuref::m * Eigen::Vector3d(ecef->x, ecef->y, ecef->z);

  enu->east = enu_(0);
  enu->north = enu_(1);
  enu->up = enu_(2);
}

void
gpsconv_ecefenu_stddev(const gps_ecef_s *ecef_stddev, gps_enu_s *enu_stddev)
{
  Eigen::Matrix3d ecef_cov_(Eigen::Matrix3d::Constant(0.));
  Eigen::Matrix3d enu_cov_;

  ecef_cov_.diagonal() <<
    ecef_stddev->x * ecef_stddev->x,
    ecef_stddev->y * ecef_stddev->y,
    ecef_stddev->z * ecef_stddev->z;

  enu_cov_ = enuref::m * ecef_cov_ * enuref::m.transpose();

  enu_stddev->east = sqrt(enu_cov_(0, 0));
  enu_stddev->north = sqrt(enu_cov_(1, 1));
  enu_stddev->up = sqrt(enu_cov_(2, 2));
}


/* --- gpsconv_llhenu ------------------------------------------------------ */

void
gpsconv_llhenu(const gps_llh_s *llh, gps_enu_s *enu)
{
  gps_ecef_s ecef;

  gpsconv_llhecef(llh, &ecef);
  gpsconv_ecefenu(&ecef, enu);
}


/* --- misc ---------------------------------------------------------------- */

/* simple Kalman filter averaging multiple measurements of the same position */
void
gpsconv_llhavg(const gps_llh_s *llh, const gps_enu_s *llh_sigma,
               gps_llh_s *avg, gps_enu_s *avg_sigma)
{
  /* observation */
  Eigen::Array3d z(
    llh->latitude, llh->longitude, llh->height);
  Eigen::Array3d R(
    llh_sigma->north * llh_sigma->north,
    llh_sigma->east * llh_sigma->east,
    llh_sigma->up * llh_sigma->up);

  /* state */
  Eigen::Array3d x(
    avg->latitude, avg->longitude, avg->height);
  Eigen::Array3d P(
    avg_sigma->north * avg_sigma->north + 1e-5 /* process noise ±1cm */,
    avg_sigma->east * avg_sigma->east + 1e-5,
    avg_sigma->up * avg_sigma->up + 1e-5);

  /* update */
  Eigen::Array3d k = P / (P + R);

  x += k * (z - x);
  P *= 1. - k;

  avg->latitude = x(0);
  avg->longitude = x(1);
  avg->height = x(2);
  avg_sigma->north = sqrt(P(0));
  avg_sigma->east = sqrt(P(1));
  avg_sigma->up = sqrt(P(2));
}
