/*
 * Copyright (c) 2018-2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Oct 18 2018
 */
#include "acgps.h"

#include <sys/time.h>

#include <err.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include <poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "codels.h"

static void *		ublox_match(struct gps_channel_s *chan);
static int		ublox_configure(struct gps_device_s *dev);
static int		ublox_unconfigure(struct gps_device_s *dev);
static int		ublox_decode(struct gps_device_s *dev,
                                struct gps_solution_s *solution,
                                sequence_octet *rtk_data);
static int		ublox_ctl(struct gps_device_s *dev,
                                enum gps_ctl_s request, ...);


const struct gps_driver_s ublox_driver = {
  .protocol = "u-blox",

  .match = ublox_match,
  .configure = ublox_configure,
  .unconfigure = ublox_unconfigure,
  .decode = ublox_decode,
  .ctl = ublox_ctl
};

struct ublox_driver_data_s {
  char model[32];		/* device model */
  double period;
};


/* --- UBX protocol -------------------------------------------------------- */

enum ublox_msgcid {
  mcid_ACK_ACK = 0x0501,
  mcid_ACK_NAK = 0x0500,

  mcid_CFG_INF = 0x0602,
  mcid_CFG_MSG = 0x0601,
  mcid_CFG_PMS = 0x0686,
  mcid_CFG_PRT = 0x0600,
  mcid_CFG_RATE = 0x0608,

  mcid_MON_MSGPP = 0x0a06,
  mcid_MON_VER = 0x0a04,

  mcid_NAV_PVT = 0x0107,

  mcid_NMEA_GGA = 0xf000,
  mcid_NMEA_GLL = 0xf001,
  mcid_NMEA_GSA = 0xf002,
  mcid_NMEA_GSV = 0xf003,
  mcid_NMEA_RMC = 0xf004,
  mcid_NMEA_VTG = 0xf005
};

struct __attribute__((packed)) ublox_ack {
  uint8_t msgclass, msgid;
};

struct __attribute__((packed)) ublox_cfg_inf {
  uint8_t proto;
  uint8_t reserved[3];
  uint8_t mask[6];
};

struct __attribute__((packed)) ublox_cfg_msg {
  uint8_t	msgclass, msgid;
  uint8_t	rate[6];
};

struct __attribute__((packed)) ublox_cfg_pms {
  uint8_t	version;
  uint8_t	powersetupvalue;
  uint16_t	period;
  uint16_t	ontime;
  uint8_t	res[2];
};

struct __attribute__((packed)) ublox_cfg_prt {
  uint8_t	portid;
  uint8_t	res[1];
  uint16_t	txready;
  union {
    struct {
      uint32_t	mode;
      uint32_t	baud;
      uint16_t	inproto;
      uint16_t	outproto;
      uint16_t	flags;
      uint8_t	res[2];
    } uart;
  };
};

struct __attribute__((packed)) ublox_cfg_rate {
  uint16_t	measrate;
  uint16_t	navrate;
  uint16_t	timeref;
};

struct __attribute__((packed)) ublox_mon_msgpp {
  uint16_t msg1[8];
  uint16_t msg2[8];
  uint16_t msg3[8];
  uint16_t msg4[8];
  uint16_t msg5[8];
  uint16_t msg6[8];
  uint32_t skipped[6];
};

struct __attribute__((packed)) ublox_mon_ver {
  char swver[30];
  char hwver[10];
  struct { char data[30]; } ext[];
};

struct __attribute__((packed)) ublox_nav_pvt {
  uint32_t	itow;
  uint16_t	year;
  uint8_t	month, day, hour, min, sec;
  uint8_t	valid;
  uint32_t	tacc;
  int32_t	nano;

  uint8_t	fixtype, flags, flags2;
  uint8_t	numsv;

  int32_t	lon, lat, height, hmsl;
  uint32_t	hacc, vacc;

  int32_t	veln, vele, veld;
  int32_t	gspeed;
  int32_t	headmot;
  uint32_t	sacc, headacc;

  uint16_t	pdop;
  uint8_t	res[6];

  int32_t	headveh;
  int16_t	magdec;
  uint16_t	magacc;
};

struct __attribute__((packed)) ublox_hdr {
  uint8_t	sync1, sync2;
  uint8_t	msgclass, msgid;
#define ublox_msgcid16(h) ((uint16_t)(((h)->msgclass << 8) | (h)->msgid))
  uint16_t	msglen;
  uint8_t	payload[];
};

struct __attribute__((packed)) ublox_msg {
  struct ublox_hdr hdr;
  union {
    struct ublox_ack ack;
    struct ublox_cfg_inf cfg_inf;
    struct ublox_cfg_msg cfg_msg;
    struct ublox_cfg_pms cfg_pms;
    struct ublox_cfg_prt cfg_prt;
    struct ublox_cfg_rate cfg_rate;
    struct ublox_mon_msgpp mon_msgpp;
    struct ublox_mon_ver mon_ver;
    struct ublox_nav_pvt nav_pvt;
  } data[];
};

#define ublox_msg(len)                                              \
  struct __attribute__((packed)) {                                  \
    struct ublox_msg	msg;                                        \
    uint8_t		payload[len];                               \
    uint16_t		cksum;                                      \
  }


static int			ublox_ack_msg(struct gps_channel_s *chan,
                                        struct ublox_msg *msg,
                                        enum ublox_msgcid mcid,
                                        uint16_t len, int timeout);
static const struct ublox_msg *	ublox_recv_msg(struct gps_channel_s *chan,
                                        const struct timeval *deadline);
static int			ublox_send_msg(struct gps_channel_s *chan,
                                        struct ublox_msg *msg,
                                        enum ublox_msgcid mcid,
                                        uint16_t len);
static uint16_t			ublox_cksum(const struct ublox_hdr *hdr);


/* --- ublox_match --------------------------------------------------------- */

static void *
ublox_match(struct gps_channel_s *chan)
{
  struct ublox_driver_data_s *ddata;
  const struct ublox_msg *msg;
  struct timeval deadline;
  int i, s;

  /* try for at most 500ms */
  gettimeofday(&deadline, NULL);
  deadline.tv_usec += 500 * 1000;
  if (deadline.tv_usec > 1000000) {
    deadline.tv_sec += deadline.tv_usec / 1000000;
    deadline.tv_usec = deadline.tv_usec % 1000000;
  }

  /* send get version message */
  ublox_msg(0) ver;
  s = ublox_send_msg(chan, &ver.msg, mcid_MON_VER, 0);
  if (s) return NULL;

  /* receive */
  do {
    msg = ublox_recv_msg(chan, &deadline);
    if (!msg) return NULL;

    /* match MON-VER message */
    if (ublox_msgcid16(&msg->hdr) != mcid_MON_VER) continue;

    /* match */
    ddata = malloc(sizeof(*ddata));
    if (!ddata) return NULL;

    for(i = 0; i < (msg->hdr.msglen - 40)/30; i++)
      warnx("ver: %s", msg->data->mon_ver.ext[i].data);

    snprintf(ddata->model, sizeof(ddata->model),
             "u-blox %s", msg->data->mon_ver.hwver);
    ddata->period = 0.;
    return ddata;

  } while(1);

  return NULL;
}


/* --- ublox_configure ----------------------------------------------------- */

static int
ublox_configure(struct gps_device_s *dev)
{
  struct ublox_driver_data_s *ddata = dev->data;
  static const struct {
    enum ublox_msgcid msgcid; uint8_t rate;
  } default_cfg_msg[] = {
    /* diable default NMEA messages */
    { mcid_NMEA_GGA, 0 },
    { mcid_NMEA_GLL, 0 },
    { mcid_NMEA_GSA, 0 },
    { mcid_NMEA_GSV, 0 },
    { mcid_NMEA_RMC, 0 },
    { mcid_NMEA_VTG, 0 },

    /* enable UBX messages */
    { mcid_NAV_PVT, 1 },
    { mcid_MON_MSGPP, 1 }
  };

  unsigned int i;
  int s;

  /* UART port configuration - resend the message thrice as it is sometimes
   * not acked */
  for(i = 0; i < 3; i++) {
    s = ublox_ctl(dev, GPSCTL_BAUD, dev->channel.baud);
    if (s) return s;
    s = ublox_ack_msg(&dev->channel, NULL, mcid_CFG_PRT, 0, 500);
    if (!s) break;
  }
  if (s) {
    warnx("u-blox UART port configuration failed");
    return s;
  }


  /* periodic messages */
  for(i = 0; i < sizeof(default_cfg_msg)/sizeof(default_cfg_msg[0]); i++) {
    ublox_msg(sizeof(struct ublox_cfg_msg)) rate = {};

    rate.msg.data->cfg_msg.msgclass = default_cfg_msg[i].msgcid >> 8;
    rate.msg.data->cfg_msg.msgid = default_cfg_msg[i].msgcid & 0xff;
    rate.msg.data->cfg_msg.rate[0] = default_cfg_msg[i].rate;

    s = ublox_ack_msg(&dev->channel, &rate.msg, mcid_CFG_MSG, 3, 500);
    if (s) {
      warnx("u-blox message 0x%04x failed", default_cfg_msg[i].msgcid);
      return s;
    }
  }


  /* information messages */
  ublox_msg(sizeof(struct ublox_cfg_inf)) inf = {};

  inf.msg.data->cfg_inf.proto = 1;	/* NMEA */
  inf.msg.data->cfg_inf.mask[1] = 0;	/* none on UART */
  s = ublox_ack_msg(
    &dev->channel, &inf.msg, mcid_CFG_INF, sizeof(struct ublox_cfg_inf), 500);
  if (s) return s;

  inf.msg.data->cfg_inf.proto = 0;	/* UBX */
  inf.msg.data->cfg_inf.mask[1] = 3;	/* INFO, WARN, ERR on UART */
  s = ublox_ack_msg(
    &dev->channel, &inf.msg, mcid_CFG_INF, sizeof(struct ublox_cfg_inf), 500);
  if (s) return s;


  /* power mode */
  ublox_msg(sizeof(struct ublox_cfg_pms)) pms = {};

  pms.msg.data->cfg_pms.powersetupvalue = 0;	/* full power */
  s = ublox_ack_msg(
    &dev->channel, &pms.msg, mcid_CFG_PMS, sizeof(struct ublox_cfg_pms), 500);
  if (s) return s;


  /* computation rate */
  ublox_msg(sizeof(struct ublox_cfg_rate)) rate = {};

  rate.msg.data->cfg_rate.measrate = 200;	/* 5Hz */
  rate.msg.data->cfg_rate.navrate = 1;		/* on every measurement */
  rate.msg.data->cfg_rate.timeref = 1;		/* GPS time sync */
  s = ublox_ack_msg(
    &dev->channel, &rate.msg, mcid_CFG_RATE, sizeof(struct ublox_cfg_rate),
    500);
  if (s) return s;
  ddata->period = 0.2;

  return 0;
}


/* --- ublox_unconfigure --------------------------------------------------- */

static int
ublox_unconfigure(struct gps_device_s *dev)
{
  (void)dev;
  return 0;
}


/* --- ublox_decode -------------------------------------------------------- */

static int
ublox_decode(struct gps_device_s *dev, struct gps_solution_s *solution,
             sequence_octet *rtk_data)
{
  (void)rtk_data;
  struct ublox_driver_data_s *ddata = dev->data;
  const struct ublox_msg *msg;
  struct timeval tv;

  while(1) {
    msg = ublox_recv_msg(&dev->channel, NULL);
    if (!msg) break;

    switch(ublox_msgcid16(&msg->hdr)) {
      case mcid_MON_MSGPP: {
        const struct ublox_mon_msgpp *pp = &msg->data->mon_msgpp;
        warnx("MSGXXX port0 %d %d %d %d %d %d %d %d",
              pp->msg2[0], pp->msg2[1], pp->msg2[2], pp->msg2[3],
              pp->msg2[4], pp->msg2[5], pp->msg2[6], pp->msg2[7]);
        warnx("MSGXXX skipped %d %d %d %d %d %d",
              pp->skipped[0], pp->skipped[1], pp->skipped[2],
              pp->skipped[3], pp->skipped[4], pp->skipped[5]);
        break;
      }

      case mcid_NAV_PVT: {
        const struct ublox_nav_pvt *pvt = &msg->data->nav_pvt;
        or_time_ts utc, ts;

        utc = gpsconv_utcmktime(&(struct tm){
            .tm_sec = pvt->sec,
              .tm_min = pvt->min,
              .tm_hour = pvt->hour,
              .tm_mday = pvt->day,
              .tm_mon = pvt->month - 1,
              .tm_year = pvt->year - 1900
              }, pvt->nano);
        ts = gpsconv_utcts(
          utc, tv, ddata->period,
          sizeof(ublox_msg(sizeof(*pvt))) * 10./*8N1*//dev->channel.baud);

        solution->utc = utc;
        solution->ts = ts;
        solution->fix = gps_fix_none;

#if 1
        warnx("itow %f", pvt->itow / 1000.);
        warnx("time %d %d %d", pvt->year, pvt->month, pvt->day);
        warnx("date %d %d %d", pvt->hour, pvt->min, pvt->sec);
        warnx("time %d.%09d",
              solution->ts.sec, solution->ts.nsec);
        warnx("valid %02x", pvt->valid);
        warnx("fixtype %02x", pvt->fixtype);
        warnx("flags %02x", pvt->flags);
        warnx("sv %02x", pvt->numsv);
        warnx("hacc %f vacc %f", 3 * pvt->hacc * 1e-3, 3 * pvt->vacc * 1e-3);
        warnx("lon %f lat %f height %f dmsl %f",
              pvt->lon * 1e-7, pvt->lat * 1e-7, pvt->height * 1e-3,
              pvt->hmsl * 1e-3 - pvt->height * 1e-3);

        gps_llh_s llh = {
          pvt->lat * 1e-7* M_PI/180.,
          pvt->lon  * 1e-7* M_PI/180.,
          pvt->height * 1e-3
        };

        gps_enu_s enu;
        gpsconv_llhenu(&llh, &enu);
        warnx("e %f n %f u %f", enu.east, enu.north, enu.up);
#endif
        break;
      }
    }
  }

  return 0;
}


/* --- ublox_ctl ----------------------------------------------------------- */

static int
ublox_ctl(struct gps_device_s *dev, enum gps_ctl_s request, ...)
{
  va_list args;
  int s;

  va_start(args, request);
  switch(request) {
    case GPSCTL_BAUD: /* set baud rate */ {
      uint32_t baud = va_arg(args, uint32_t);

      ublox_msg(sizeof(struct ublox_cfg_prt)) prt = {};
      warnx("XXX baud %d", baud);

      /* serial config */
      prt.msg.data->cfg_prt.portid = 1;			/* UART 1 */
      prt.msg.data->cfg_prt.txready = 0;		/* tx pin disabled */
      prt.msg.data->cfg_prt.uart.mode = (3<<6)|(4<<9);	/* 8N1 */
      prt.msg.data->cfg_prt.uart.baud = baud;		/* baud */
      prt.msg.data->cfg_prt.uart.inproto = 5;		/* UBX/RTCM2 */
      prt.msg.data->cfg_prt.uart.outproto = 1;		/* UBX */
      prt.msg.data->cfg_prt.uart.flags = 0;		/* 1.5s timeout */

      s = ublox_send_msg(&dev->channel, &prt.msg,
                         mcid_CFG_PRT, sizeof(struct ublox_cfg_prt));
      break;
    }

    case GPSCTL_INRTCM: /* feed RTK data */ {
      /* uint16_t type = */ va_arg(args, unsigned int);
      uint8_t *data = va_arg(args, uint8_t *);
      size_t len = va_arg(args, size_t);

      if (len == 0) { s = 0; break; }

      s = gpstty_send(&dev->channel, data, len);
      break;
    }

    default: s = 1;
  }
  va_end(args);

  return s;
}


/* --- ublox_ack_msg ------------------------------------------------------- */

static int
ublox_ack_msg(struct gps_channel_s *chan, struct ublox_msg *msg,
              enum ublox_msgcid mcid, uint16_t len, int timeout)
{
  const struct ublox_msg *ack;
  struct timeval deadline;
  int s;

  gettimeofday(&deadline, NULL);
  deadline.tv_usec += timeout * 1000;
  if (deadline.tv_usec > 1000000) {
    deadline.tv_sec += deadline.tv_usec / 1000000;
    deadline.tv_usec = deadline.tv_usec % 1000000;
  }

  if (msg) {
    s = ublox_send_msg(chan, msg, mcid, len);
    if (s) return s;
  }

  /* ignore all messages until corresponding ACK */
  while(1) {
    ack = ublox_recv_msg(chan, &deadline);
    if (!ack) return -1;
    if (ublox_msgcid16(&ack->hdr) != mcid_ACK_ACK &&
        ublox_msgcid16(&ack->hdr) != mcid_ACK_NAK) continue;
    if (ublox_msgcid16(&ack->data->ack) == mcid)
      break;
  }

  return ublox_msgcid16(&ack->hdr) != mcid_ACK_ACK;
}


/* --- ublox_recv_msg ------------------------------------------------------ */

static const struct ublox_msg *
ublox_recv_msg(struct gps_channel_s *chan, const struct timeval *deadline)
{
  static const struct ublox_hdr expect = {
    .sync1 = 0xb5, .sync2 = 0x62
  };

  const struct ublox_hdr *hdr;
  size_t len, req;
  ssize_t n;

  do {
    /* peek one header */
    hdr = gpstty_data(chan, 0, &len);
    if (len < sizeof(*hdr)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break; /* err or not enough data */
      continue;
    }

    /* match sync bytes */
    if (hdr->sync1 != expect.sync1 || hdr->sync2 != expect.sync2) {
      gpstty_data(chan, 1, NULL); /* skip 1 byte */
      continue;
    }

    /* get enough data for the whole message */
    const ublox_msg(hdr->msglen) *msg;

    for(req = sizeof(*msg); len < req; gpstty_data(chan, 0, &len)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break;
    }

    msg = gpstty_data(chan, 0, &len);
    if (len < req) break; /* err or not enough data */

    /* check crc */
    if (msg->cksum != ublox_cksum(&msg->msg.hdr)) {
      size_t s;

      warnx("u-blox message 0x%02x%02x with bad checksum",
            msg->msg.hdr.msgclass, msg->msg.hdr.msgid);

      for(s = 0; s < req; s++) warnx("XXX         0x%02x (%c) [%zu]",
                                              ((char *)msg)[s], ((char
                                              *)msg)[s] > 32 ? ((char *)msg)[s]
                                              : ' ', s);

      gpstty_data(chan, 2, NULL); /* skip 2 bytes */
      continue;
    }

#if 0
    warnx("XXX message %02x%02x", msg->msg.hdr.msgclass, msg->msg.hdr.msgid);
#endif
    gpstty_data(chan, req, NULL);
    return &msg->msg;

  } while(1);

  return NULL;
}


/* --- ublox_send_msg ------------------------------------------------------ */

static int
ublox_send_msg(struct gps_channel_s *chan, struct ublox_msg *msg,
               enum ublox_msgcid mcid, uint16_t len)
{
  msg->hdr = (struct ublox_hdr){
    .sync1 = 0xb5, .sync2 = 0x62,
    .msgclass = mcid >> 8,
    .msgid = mcid & 0xff,
    .msglen = len
  };
  ((ublox_msg(len) *)msg)->cksum = ublox_cksum(&msg->hdr);

  return gpstty_send(chan, msg, sizeof(ublox_msg(len)));
}


/* --- ublox_cksum --------------------------------------------------------- */

static uint16_t
ublox_cksum(const struct ublox_hdr *hdr)
{
  uint16_t len = hdr->msglen;
  uint8_t cksum_a, cksum_b;
  const uint8_t *p;

  cksum_a  = hdr->msgclass;      cksum_b  = cksum_a;
  cksum_a += hdr->msgid;         cksum_b += cksum_a;
  cksum_a += hdr->msglen & 0xff; cksum_b += cksum_a;
  cksum_a += hdr->msglen >> 8;   cksum_b += cksum_a;

  for(p = hdr->payload; len--; p++) {
    cksum_a += *p;               cksum_b += cksum_a;
  }

  return (cksum_b << 8) | cksum_a;
}
